import sklearn
from patch.statsd_config import DEFAULT_STATSD_CLIENT as STATSD_CLIENT
from sklearn import model_selection
from sklearn import linear_model


_LogisticRegression = linear_model.LogisticRegression

_train_test_split = model_selection.train_test_split

sklearn_string = "sklearn.."
logreg_string = "logistic_regression.."


def train_test_split_patched(*args, **kwargs):
    """
    New train_test_split function
    """
    with STATSD_CLIENT.timer(f"{sklearn_string}train_test_split"):
        return _train_test_split(*args, **kwargs)


class PatchedLogisticRegression(_LogisticRegression):
    def predict_proba(self, X):
        with STATSD_CLIENT.timer(f"{sklearn_string}{logreg_string}predict_proba"):
            return super(PatchedLogisticRegression, self).predict_proba(X) 
    
    def fit(self, X, y, sample_weight=None):
        with STATSD_CLIENT.timer(f"{sklearn_string}{logreg_string}fit"):
            return super(PatchedLogisticRegression, self).fit(X, y, sample_weight) 


def patch():
    model_selection.train_test_split = train_test_split_patched
    linear_model.LogisticRegression = PatchedLogisticRegression



from statsd.client import timer
from datetime import timedelta
from statsd.client import udp
from tracing import DefaultContext
import time
from tracing.config import SPAN_RECORD


# Use timer that's not susceptible to time of day adjustments.
try:
    # perf_counter is only present on Py3.3+
    from time import perf_counter as time_now
except ImportError:
    # fall back to using time
    from time import time as time_now

_StatsClient = udp.StatsClient
_Timer = timer.Timer


class PatchedStatsClient(_StatsClient):
    """
    PatchStatsClient that returns custom JSON format containing trace information
    """


    def timer(self, stat, rate=1):
        return PatchedTimer(self, stat, rate)
    
    def parametrized(self, stat, rate=1):
        return ParametrizedPatchedTimer(self, stat, rate)

    def timing(self, stat, start_time, delta, parent, self_id, rate=1, meta={}):
        """
        Send new timing information.
        `delta` can be either a number of milliseconds or a timedelta.
        """

        if isinstance(delta, timedelta):
            # Convert timedelta to number of microseconds
            delta_microseconds = int(delta.total_milliseconds() * 1000.0)
        else:
            delta_microseconds = int(delta * 1000.0)
       
        span_record = SPAN_RECORD(
            parent_id=parent,
            span_id=self_id,
            start_time=start_time,
            trace_id=DefaultContext.trace_id,
            service=DefaultContext.service,
            operation=stat,
            end_time=start_time+delta_microseconds,
            children=DefaultContext.child_map[self_id],
            meta=meta,
        )

        DefaultContext.add_span_to_trace_buffer(span_record=span_record)


class PatchedTimer(_Timer):
    """
    Patched Timer that keeps track of parent state and collects information that
    allows the Stats Client to return the JSON format
    """

    def __call__(self, f):
        """Thread-safe timing function decorator."""

        @timer.safe_wraps(f)
        def _wrapped(*args, **kwargs):
            """
            Patching the Timer when called by decorator
            :param args:
            :param kwargs:
            :return:
            """
            parent = DefaultContext.get_current_span_from_parent_tree()
            DefaultContext.add_new_span_to_parent_tree()
            self_id = DefaultContext.get_current_span_from_parent_tree()
            duration_begin = time_now()
            start_time = int(time.time() * (10 ** 6))
            try:
                return f(*args, **kwargs)
            finally:
                elapsed_time_ms = 1000.0 * (time_now() - duration_begin)
                self_id = DefaultContext.get_current_span_from_parent_tree()
                DefaultContext.remove_current_span_from_parent_tree()
                self.client.timing(
                    self.stat,
                    start_time,
                    elapsed_time_ms,
                    parent,
                    self_id,
                    self.rate,
                )

        return _wrapped

    """
    The start & send functions below work when the Timer is called with a "with"
    statement. 
    """

    def start(self):
        # statsd.Timer class members
        self.ms = None
        self._sent = False
        self._start_time = time_now()
        
        # extensions
        self.__clock_start_time = int(time.time() * (10 ** 6))
        self.__parent = DefaultContext.get_current_span_from_parent_tree()
        DefaultContext.add_new_span_to_parent_tree()
        return self

    def send(self):
        if self.ms is None:
            raise RuntimeError("No data recorded.")
        if self._sent:
            raise RuntimeError("Already sent data.")
        self._sent = True
        self_id = DefaultContext.get_current_span_from_parent_tree()
        DefaultContext.remove_current_span_from_parent_tree()
        self.client.timing(
            self.stat,
            self.__clock_start_time,
            self.ms,
            self.__parent,
            self_id,
            self.rate,
        )


class ParametrizedPatchedTimer(_Timer):
    """
    Patched Timer that keeps track of parent state and collects information that
    allows the Stats Client to return the JSON format
    """

    def __call__(self, f):
        """Thread-safe timing function decorator."""

        @timer.safe_wraps(f)
        def _wrapped(*args, **kwargs):
            """
            Patching the Timer when called by decorator
            :param args:
            :param kwargs:
            :return:
            """
            parent = DefaultContext.get_current_span_from_parent_tree()
            DefaultContext.add_new_span_to_parent_tree()
            self_id = DefaultContext.get_current_span_from_parent_tree()
            duration_begin = time_now()
            start_time = int(time.time() * (10 ** 6))

            arg_params = args
            kwarg_params = kwargs

            try:
                return f(*args, **kwargs)
            finally:
                elapsed_time_ms = 1000.0 * (time_now() - duration_begin)
                self_id = DefaultContext.get_current_span_from_parent_tree()
                DefaultContext.remove_current_span_from_parent_tree()
                self.client.timing(
                    self.stat,
                    start_time,
                    elapsed_time_ms,
                    parent,
                    self_id,
                    self.rate,
                    meta={"args": arg_params, "kwargs": kwarg_params}
                )

        return _wrapped

    """
    The start & send functions below work when the Timer is called with a "with"
    statement. 
    """

    def start(self):
        # statsd.Timer class members
        self.ms = None
        self._sent = False
        self._start_time = time_now()
        
        # extensions
        self.__clock_start_time = int(time.time() * (10 ** 6))
        self.__parent = DefaultContext.get_current_span_from_parent_tree()
        DefaultContext.add_new_span_to_parent_tree()
        return self

    def send(self):
        if self.ms is None:
            raise RuntimeError("No data recorded.")
        if self._sent:
            raise RuntimeError("Already sent data.")
        self._sent = True
        self_id = DefaultContext.get_current_span_from_parent_tree()
        DefaultContext.remove_current_span_from_parent_tree()
        self.client.timing(
            self.stat,
            self.__clock_start_time,
            self.ms,
            self.__parent,
            self_id,
            self.rate,
        )



from collections import namedtuple

DEFAULT_AGENT_URL = "http://127.0.0.1:8000/api/traces"
DEFAULT_REQUEST_BUFFER_SIZE = 20

SPAN_RECORD = namedtuple(
    "SpanRecord",
    "parent_id, span_id, start_time, trace_id, service, operation, end_time, children, meta",
)


def convert_span_record_to_key_value(span_record: SPAN_RECORD):
    """
    This is a separate function only so we can modify this code if asdict() is deprecated.
    :param span_record:
    :return:
    """
    return span_record._asdict()



from collections import deque
import uuid
import tracing.config as config
import requests
import json
import random


class LocalContext(object):
    """
    Class that stores context about the current nested set of traces
    It keeps track of parent-child relationships

    If a span exists, it has a corresponding empty child map
    """

    def __init__(self, trace_id=None, span_id=None, service="Default"):
        if span_id is None:
            span_id = random.getrandbits(32)

        if trace_id is None:
            trace_id = random.getrandbits(32)

        self.trace_id = trace_id
        self.child_map = {}
        self.spans_by_id = {}
        self.ordered_trace_deque = deque([span_id])
        self.child_map[span_id] = []
        self.service = service
        self.trace_buffer = {self.trace_id: []}

    def get_current_span_from_parent_tree(self):
        """
        Returns the trace that is currently at the lowest level
        in the dependency tree
        :return:
        """
        return self.ordered_trace_deque[-1]

    def remove_current_span_from_parent_tree(self):
        """
        Removes a trace from the lowest level of the dependency tree.
        :return:
        """
        self.ordered_trace_deque.pop()

    def add_new_span_to_parent_tree(self, new_trace=None):
        if new_trace is None:
            new_trace = random.getrandbits(32)
        self.child_map[new_trace] = []
        self.child_map[self.get_current_span_from_parent_tree()].append(new_trace)
        self.ordered_trace_deque.append(new_trace)

    def add_span_to_trace_buffer(self, span_record: config.SPAN_RECORD):
        span_as_map = config.convert_span_record_to_key_value(span_record)
        self.trace_buffer[self.trace_id].append(span_as_map)
        self.spans_by_id[span_as_map["span_id"]] = span_as_map

        if len(self.trace_buffer[self.trace_id]) >= config.DEFAULT_REQUEST_BUFFER_SIZE:
            self.flush_buffer_to_agent()

    def _get_fully_described_child(self, child):
        span_as_map = self.spans_by_id[child]
        span_as_map["id"] = span_as_map["span_id"]
        span_as_map["children"] = [self._get_fully_described_child(child) for child in span_as_map["children"]]

        return span_as_map

    def _get_fully_described_trace(self, trace):
        # find the root span
        root = None

        for span in trace:
            if span["parent_id"] == 0:
                root = span
                break
        
        if not root:
            raise RuntimeException("no root for trace_id %s" % trace['trace_id'])

        root["children"] = [self._get_fully_described_child(child) for child in root["children"]]
        root["meta"] = {"layer": "sigmoid", "neurons": 4, "next_layer": "curvilinear"}
        root["id"] = root["span_id"]

        return root


    def flush_buffer_to_agent(self):
        # Trace payload should look like
        # { "traces": [ { "id': "xyz", "trace": {} }, ...  ] }


        trace_list = {"traces": []}

        for trace_id, trace in self.trace_buffer.items():
            trace_list["traces"].append(
                {"id": trace_id, "trace": self._get_fully_described_trace(trace)}
            )
        
        print(json.dumps(trace_list))
        requests.post(url=config.DEFAULT_AGENT_URL, json=trace_list)
        self.trace_buffer.clear()
        self.trace_buffer[self.trace_id] = []



from distutils.core import setup

setup(
    name='athena-run',
    version='0.1.2dev',
    packages=['tracing','patch'],
    license='Creative Commons Attribution-Noncommercial-Share Alike license',
    long_description="ML Tracing Tools",
)



import pandas
from pandas import DataFrame

from patch.statsd_config import DEFAULT_STATSD_CLIENT as STATSD_CLIENT

_read_parquet = pandas.read_parquet
_to_parquet = DataFrame.to_parquet
_merge = pandas.merge
_groupby = DataFrame.groupby

pandas_string = "pandas.."


def read_parquet_patched(*args, **kwargs):
    """
    New read_parquet function   
    """
    with STATSD_CLIENT.timer(f"{pandas_string}read_parquet"):
        return _read_parquet(*args, **kwargs)


def to_parquet_patched(*args, **kwargs):
    """
    New to_parquet function   
    """
    with STATSD_CLIENT.timer(f"{pandas_string}to_parquet"):
        return _to_parquet(*args, **kwargs)


def merge_patched(*args, **kwargs):
    """
    New merge function   
    """
    with STATSD_CLIENT.timer(f"{pandas_string}merge"):
        return _merge(*args, **kwargs)


def dataframe_groupby_patched(*args, **kwargs):
    """
    New read parquet function   
    """
    with STATSD_CLIENT.timer(f"{pandas_string}df_groupby"):
        return _groupby(*args, **kwargs)


def patch():
    pandas.read_parquet = read_parquet_patched
    DataFrame.to_parquet = to_parquet_patched
    pandas.merge = merge_patched
    DataFrame.groupby = dataframe_groupby_patched
